import itertools
import logging
import os
import re
from collections import OrderedDict
from io import UnsupportedOperation
from typing import Dict, Any, List, Optional, Set
from typing import Tuple
import pandas as pd

import fsspec
import parse
from fsspec import AbstractFileSystem

log = logging.getLogger('rawfiloc')

# ---------
# Constants
# ---------
_improbable_string    = "o=NvZ_ps$"
_re_improbable_string = re.compile(r'(o=NvZ_ps\$)')
_re_natural           = re.compile(r"(\d+)")
_re_path_placeholder  = re.compile(r'({[^}]+})')


# -------
# Helpers
# -------
def sort_natural(li: List[str]) -> List[str]:
    # todo: support float too
    return sorted(li, key=lambda s: [int(part) if part.isdigit() else part for part in _re_natural.split(s)])


def mix_dicts_and_coerce(dict1, dict2):
    if isinstance(dict1, pd.Series):
        dict1 = dict1.to_dict()
    if isinstance(dict2, pd.Series):
        dict2 = dict2.to_dict()

    dict2 = None if len(dict2) == 0 else dict2
    if dict1 and dict2:
        return OrderedDict([dict1, dict2])
    elif dict1:
        return dict1
    elif dict2:
        return dict2
    else:
        return {}


# -------------------
# Base Class RawFiloc
# -------------------
class FilocIO:
    def __init__(self, locpath: str, writable=False) -> None:
        super().__init__()
        self.locpath     = locpath
        self.writable    = writable
        path_elts        = _re_path_placeholder.split("_" + locpath)  # dummy _ ensures that first elt is not a placeholder
        path_elts[0]     = path_elts[0][1:]  # removes _
        placeholders     = path_elts[1::2]
        valid_path       = "".join([e1+e2 for e1, e2 in zip(path_elts[::2], itertools.repeat(_improbable_string))])
        valid_file       = fsspec.open(valid_path)
        self.fs          = valid_file.fs           # type: AbstractFileSystem
        norm_elts        = _re_improbable_string.split("_" + valid_file.path)  # dummy _ ensures that first elt is not a placeholder
        norm_elts[0]     = norm_elts[0][1:]  # removes _
        norm_path        = "".join([e1+e2 for e1, e2 in itertools.zip_longest(norm_elts[::2], placeholders, fillvalue="")])
        self.locpath     = norm_path  # type: str
        self.path_parser = parse.compile(self.locpath)  # type: parse.Parser
        self.root_folder = self.fs.sep.join((norm_elts[0] + "AAA").split(self.fs.sep)[:-1])  # AAA ensures that the path ends with a file name/folder name to remove
        # noinspection PyProtectedMember
        self.path_props  = set(self.path_parser._named_fields)  # type: Set[str]

    # noinspection PyDefaultArgument
    def get_path_properties(self, path: str) -> Dict[str, Any]:
        try:
            return self.path_parser.parse(path).named
        except Exception as e:
            raise ValueError(f'Could not parse {path} with {self.locpath} parser: {e}')

    def get_path(self, path_props : Optional[Dict[str, Any]] = None, **path_props_kwargs : Any) -> str:
        path_props = mix_dicts_and_coerce(path_props, path_props_kwargs)
        undefined_keys = self.path_props - set(path_props)
        if len(undefined_keys) > 0:
            raise ValueError('Required props undefined: {}. Provided: {}'.format(undefined_keys, path_props))
        return self.locpath.format(**path_props)  # result should be normalized, because locpath is

    def get_glob_path(self, path_props : Optional[Dict[str, Any]] = None, **path_props_kwargs : Any) -> str:
        path_props = mix_dicts_and_coerce(path_props, path_props_kwargs)
        provided_keys = set(path_props)
        undefined_keys = self.path_props - provided_keys
        defined_keys = self.path_props - undefined_keys

        path_values = OrderedDict()
        path_values.update({(k, path_props[k]) for k in defined_keys})

        glob_path = self.locpath
        for undefined_key in undefined_keys:
            glob_path = re.sub(r'{' + undefined_key + r'(?::[^}]*)?}', '?*', glob_path)

        # finally format
        glob_path = glob_path.format(**path_values)
        return glob_path  # result should be normalized, because locpath is

    def find_paths(self, path_props : Optional[Dict[str, Any]] = None, **path_props_kwargs : Any) -> List[str]:
        path_props = mix_dicts_and_coerce(path_props, path_props_kwargs)
        paths = self.fs.glob(self.get_glob_path(path_props))
        return sort_natural(paths)

    def find_paths_and_path_props(self, path_props : Optional[Dict[str, Any]] = None, **path_props_kwargs : Any) -> List[Tuple[str, List[str]]]:
        path_props = mix_dicts_and_coerce(path_props, path_props_kwargs)
        paths = self.find_paths(path_props)
        return [(p, self.get_path_properties(p)) for p in paths]

    def exists(self, path_props : Optional[Dict[str, Any]] = None, **path_props_kwargs : Any) -> bool:
        path_props = mix_dicts_and_coerce(path_props, path_props_kwargs)
        return self.fs.exists(self.get_path(path_props))

    def open(self, path_props : Dict[str, Any], mode="rb", block_size=None, cache_options=None, **kwargs):
        is_writing = len(set(mode) & set("wa+")) > 0
        if is_writing and not self.writable:
            raise UnsupportedOperation('this filoc is not writable. Set writable flag to True to enable writing')

        path = self.get_path(path_props)
        dirname = os.path.dirname(path)

        if is_writing:
            self.fs.makedirs(dirname, exist_ok=True)

        return self.fs.open(path, mode, block_size, cache_options, **kwargs)

    # noinspection PyDefaultArgument
    def delete(self, path_props : Optional[Dict[str, Any]] = {}, dry_run= False):
        if not self.writable:
            raise UnsupportedOperation('this filoc is not writable. Set writable flag to True to enable deleting')

        path_to_delete = self.find_paths(path_props)
        dry_run_log_prefix = '(dry_run) ' if dry_run else ''
        log.info(f'{dry_run_log_prefix}Deleting {len(path_to_delete)} files with path_props "{path_props}"')
        for path in path_to_delete:
            log.info(f'{dry_run_log_prefix}Deleting "{path}"')
            if dry_run:
                continue
            self.fs.delete(path)
        log.info(f'{dry_run_log_prefix}Deleted {len(path_to_delete)} files with path_props "{path_props}"')

    def __eq__(self, other):
        if other is not self:
            return False
        return self.locpath == other.locpath

    def __hash__(self):
        return self.locpath.__hash__()



import logging

from .contract import TContents, PropsList, TContent, FrontendContract

log = logging.getLogger('filoc')


class JsonFrontend(FrontendContract):
    def read_content(self, props_list: PropsList) -> TContent:
        assert len(props_list) == 1, f'Expected singleton, got {len(props_list)} items to convert to content'
        return props_list[0]

    def read_contents(self, props_list: PropsList) -> TContents:
        return props_list

    def write_content(self, content: TContent) -> PropsList:
        return [content]

    def write_contents(self, contents: TContents) -> PropsList:
        return contents



import json
import copy
import logging
from collections import OrderedDict
from typing import Dict, Any, List, Union

from orderedset import OrderedSet

from filoc.contract import PropsList

log = logging.getLogger('filoc_utils')

_missing_key = object()


def filter_and_coerce_loaded_file_content(path, file_content, constraints, is_singleton):
    # validate file_content and coerce to list
    if is_singleton:
        if not isinstance(file_content, dict):
            raise ValueError(f"Invalid type for loaded file '{path}'. Expected dict, got {type(file_content).__name__}")
        unfiltered_result = [file_content]
    else:
        if not isinstance(file_content, list):
            raise ValueError(f"Invalid type for loaded file '{path}'. Expected list, got {type(file_content).__name__}")
        unfiltered_result = file_content

    # filter
    result = []
    for row in unfiltered_result:
        keep = True
        for constraint_key, constraint_value in constraints.items():
            if constraint_key not in row:
                continue  # constraint does not apply to this row
            if row[constraint_key] != constraint_value:
                keep = False
                break
        if keep:
            result.append(row)
    return result


def coerce_file_content_to_write(path, props_list : PropsList, is_singleton : bool) -> Union[dict, list]:
    # validate file_content and coerce to list
    if is_singleton:
        if len(props_list) == 1:
            return props_list[0]
        elif len(props_list) == 0:
            raise ValueError('props_list is empty')
        else:
            first_item = props_list[0]
            for idx, item in enumerate(props_list[1:]):
                if first_item != item:
                    raise ValueError(f'Trying to save {len(props_list)} props to singleton file {path} with different values:\nItem #0: {json.dumps(first_item)}\nItem #{idx + 1}: {json.dumps(item)}')
    else:
        return props_list


# ---------------------
# Pivot Helpers
# ---------------------
def merge_tables(table_by_name : Dict[str, List[Dict[str, Any]]], join_key_names : List[str], separator : str, join_level_name : str):
    resulting_pivot = None
    for table_name, table_values in table_by_name.items():
        pi = _pivot(table_values, join_key_names, f'{table_name}{separator}')
        if resulting_pivot is None:
            resulting_pivot = pi
            continue
        else:
            _merge_pivots_recursive(resulting_pivot, pi, join_key_names)

    index_prefix = f"{join_level_name}{separator}"
    return _unpivot(resulting_pivot, join_key_names, index_prefix)


def _merge_pivots_recursive(pi1 : dict, pi2 : dict, remaining_key_names : List[str]):
    if len(remaining_key_names) == 0:
        pi1.update(pi2)
    else:
        cp1 = pi1.pop(_missing_key) if _missing_key in pi1 else OrderedDict()
        cp2 = pi2.pop(_missing_key) if _missing_key in pi2 else OrderedDict()
        all_keys = OrderedSet(pi1) | OrderedSet(pi2)

        if len(all_keys) == 0:
            _merge_pivots_recursive(cp1, cp2, remaining_key_names[1:])
            pi1[_missing_key] = cp1
        else:
            for k in all_keys:
                v1 = pi1[k] if k in pi1 else copy.deepcopy(cp1)
                v2 = pi2[k] if k in pi2 else copy.deepcopy(cp2)
                _merge_pivots_recursive(v1, v2, remaining_key_names[1:])
                pi1[k] = v1


def _pivot(table_values : List[Dict[str, Any]], key_names : List[str], prefix : str):
    result = OrderedDict()
    for item in table_values:
        path = []
        curr_level = result
        for key_name in key_names:
            key = item.pop(key_name, _missing_key)
            path.append(key)
            if key not in curr_level:
                curr_level[key] = OrderedDict()
            curr_level = curr_level[key]

        for (k, v) in item.items():
            curr_level[prefix + k] = v

    return result


def _unpivot(pivot, key_names: List[str], index_prefix : str):
    result = []
    _unpivot_recursive(pivot, OrderedDict(), key_names, index_prefix, result)
    return result


def _unpivot_recursive(pivot_node: Dict[str, Any], current_index: Dict[str, Any], remaining_key_names: List[str], index_prefix : str, result):
    if len(remaining_key_names) == 0:
        current_index.update(pivot_node)
        result.append(current_index)
    else:
        key_name = remaining_key_names[0]
        for key, sub_pivot in pivot_node.items():
            sub_index = current_index.copy()
            sub_index[index_prefix + key_name] = key
            _unpivot_recursive(sub_pivot, sub_index, remaining_key_names[1:], index_prefix, result)



import json
import shutil
import tempfile
import unittest

# noinspection DuplicatedCode
from filoc import filoc_json, FilocIO


class TestMultiloc_TwoLevels(unittest.TestCase):
    def setUp(self):
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')
        self.node_loc = filoc_json(f'{self.test_dir}/{{node_id:d}}/node.json', writable=True)
        self.leaf_loc = filoc_json(f'{self.test_dir}/{{node_id:d}}/{{leaf_id:d}}/leaf.json', writable=True)
        self.sut_loc = filoc_json({
            'node' : self.node_loc,
            'leaf' : self.leaf_loc,
        })

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_1(self):
        self.node_loc.write_content( {'node_id': 1, 'value' : 'A'})
        self.node_loc.write_content( {'node_id': 2, 'value' : 'B'})
        self.leaf_loc.write_content( {'node_id': 1, 'leaf_id': 1, 'value' : 0.1})
        self.leaf_loc.write_content( {'node_id': 1, 'leaf_id': 2, 'value' : 0.2})
        self.leaf_loc.write_content( {'node_id': 2, 'leaf_id': 1, 'value' : 1.1})
        self.leaf_loc.write_content( {'node_id': 2, 'leaf_id': 2, 'value' : 1.2})

        result = self.sut_loc.read_contents()

        result_tree = pivot(result, "index.node_id", "index.leaf_id")
        result_tree_txt = json.dumps(result_tree, sort_keys=True)
        print(result_tree_txt)

        self.maxDiff = 2000
        self.assertEquals(result_tree_txt, 
            '{'
            '"1": {'
              '"1": {"index.leaf_id": 1, "index.node_id": 1, "leaf.value": 0.1, "node.value": "A"}, '
              '"2": {"index.leaf_id": 2, "index.node_id": 1, "leaf.value": 0.2, "node.value": "A"}'
            '}, '
            '"2": {'
              '"1": {"index.leaf_id": 1, "index.node_id": 2, "leaf.value": 1.1, "node.value": "B"}, '
              '"2": {"index.leaf_id": 2, "index.node_id": 2, "leaf.value": 1.2, "node.value": "B"}'
              '}'
            '}'
        )

def pivot(l, prop1, prop2):
    result = {}
    for item in l:
        value1 = item[prop1]
        node1 = result.get(value1, {})
        result[value1] = node1

        value2 = item[prop2]
        node1[value2] = item

    return result

if __name__ == '__main__':
    unittest.main()



from collections import OrderedDict
from typing import Union, Dict, Optional, List, Any, Iterable

from pandas import Series, DataFrame

from .core import FilocComposite, Filoc
from .filoc_io import FilocIO

__version__ = '0.0.6'

from .contract import PresetFrontends, FilocContract, BackendContract, FrontendContract, PresetBackends


_default_frontend        = 'pandas'  
_default_backend         = 'json'
_default_singleton       =  True
_default_writable        =  False
_default_cache_locpath   =  None
_default_timestamp_col   =  None
_default_join_keys       =  None
_default_join_level_name =  'index'
_default_join_separator  =  '.'

def _get_frontend(frontend : Union[PresetFrontends, FrontendContract]):
    if frontend == 'json':
        from .frontend_json import JsonFrontend
        return JsonFrontend()
    elif frontend == 'pandas':
        from .frontend_pandas import PandasFrontend
        return PandasFrontend()
    else:
        return frontend


def _get_backend(backend : Union[PresetBackends, BackendContract], is_singleton : bool):
    if backend == 'csv':
        from .backend_csv import CsvBackend
        return CsvBackend(is_singleton)
    elif backend == 'json':
        from .backend_json import JsonBackend
        return JsonBackend(is_singleton)
    elif backend == 'pickle':
        from .backend_pickle import PickleBackend
        return PickleBackend(is_singleton)
    elif backend == 'yaml':
        from .backend_yaml import YamlBackend
        return YamlBackend(is_singleton)
    else:
        return backend


def filoc(
        locpath          : Union[str, Dict[str, str], Dict[str, FilocContract]],
        frontend         : Union[PresetFrontends, FrontendContract]            = _default_frontend,
        backend          : Union[PresetBackends, BackendContract]              = _default_backend,
        singleton        : bool                                                = _default_singleton,
        writable         : Optional[bool]                                      = _default_writable,
        cache_locpath    : Optional[str]                                       = _default_cache_locpath,
        timestamp_col    : Optional[str]                                       = _default_timestamp_col,
        join_keys        : Union[Iterable[str], Dict[Dict, Iterable[str]]]     = _default_join_keys,
        join_level_name  : Optional[str]                                       = _default_join_level_name,
        join_separator   : Optional[str]                                       = _default_join_separator,
) -> FilocContract:
    frontend_impl = _get_frontend(frontend)
    backend_impl  = _get_backend(backend, singleton)

    if isinstance(locpath, dict):
        filoc_by_name = OrderedDict()
        for sub_filoc_name, sub_filoc in locpath.items():
            if isinstance(sub_filoc, str):
                sub_locpath = sub_filoc
                filoc_instance = filoc(
                    locpath            = sub_locpath        ,
                    frontend           = frontend           ,
                    backend            = backend            ,
                    singleton          = singleton          ,
                    writable           = writable           ,
                    cache_locpath      = None               ,  # Remark: cache is not forwarded to sub-filocs
                    timestamp_col      = timestamp_col      ,
                    join_keys          = join_keys          ,
                    join_level_name    = join_level_name    ,
                    join_separator     = join_separator     ,
                )
            else:
                filoc_instance = sub_filoc
            filoc_by_name[sub_filoc_name] = filoc_instance

        return FilocComposite(
            filoc_by_name           = filoc_by_name,
            frontend                = frontend_impl,
            join_keys_by_filoc_name = join_keys,
            join_level_name         = join_level_name,
            join_separator          = join_separator,
        )
    elif isinstance(locpath, str):
        return Filoc(
            locpath            = locpath,
            writable           = writable,
            frontend           = frontend_impl,
            backend            = backend_impl,
            cache_locpath      = cache_locpath,
            timestamp_col      = timestamp_col
        )
    else:
        raise ValueError(f'locpath must be an instance of str or dict, but is {type(locpath)}')


def filoc_json(
        locpath            : Union[str, Dict[str, str], Dict[str, FilocContract]] = None,
        backend            : Union[PresetBackends, BackendContract]               = 'json',
        singleton          : bool                                                 = True,
        writable           : Optional[bool]                                       = False,
        cache_locpath      : Optional[str]                                        = None,
        timestamp_col      : Optional[str]                                        = None,
        join_keys          : Union[Iterable[str], Dict[Dict, Iterable[str]]]      = None,
        join_level_name    : Optional[str]                                        = 'index',
        join_separator     : Optional[str]                                        = '.',
        **locpath_kwargs   : Union[str, Dict[str, str], Dict[str, FilocContract]]
) -> FilocContract[Dict[str, Any], List[Dict[str, Any]]]:
    """ Same as filoc(), but with typed return value to improve IDE support """
    return filoc(
        locpath            = locpath            ,
        frontend           = _get_frontend('json'),
        backend            = backend            ,
        singleton          = singleton          ,
        writable           = writable           ,
        cache_locpath      = cache_locpath      ,
        timestamp_col      = timestamp_col      ,
        join_keys          = join_keys          ,
        join_level_name    = join_level_name    ,
        join_separator     = join_separator     ,
        **locpath_kwargs
    )


def filoc_pandas(
        locpath            : Union[str, Dict[str, str], Dict[str, FilocContract]] = None,
        backend            : Union[PresetBackends, BackendContract]               = 'json',
        singleton          : bool                                                 = True,
        writable           : Optional[bool]                                       = False,
        cache_locpath      : Optional[str]                                        = None,
        timestamp_col      : Optional[str]                                        = None,
        join_keys          : Union[Iterable[str], Dict[Dict, Iterable[str]]]      = None,
        join_level_name    : Optional[str]                                        = 'index',
        join_separator     : Optional[str]                                        = '.',
        **locpath_kwargs   : Union[str, Dict[str, str], Dict[str, FilocContract]]
) -> FilocContract[Series, DataFrame]:
    """ Same as filoc(), but with typed return value to improve IDE support """
    return filoc(
        locpath            = locpath            ,
        frontend           = _get_frontend('pandas'),
        backend            = backend            ,
        singleton          = singleton          ,
        writable           = writable           ,
        cache_locpath      = cache_locpath ,
        timestamp_col      = timestamp_col      ,
        join_keys          = join_keys          ,
        join_level_name    = join_level_name    ,
        join_separator     = join_separator     ,
        **locpath_kwargs
    )


__all__     = [
    'filoc',
    'FilocIO',
    'filoc_json',
    'filoc_pandas',
]



# -------------
# Types aliases
# -------------
from abc import ABC
from typing import TypeVar, Literal, Dict, Any, List, Generic, Optional

from fsspec import AbstractFileSystem

TContent             = TypeVar('TContent')
TContents            = TypeVar('TContents')
PresetFrontends      = Literal['json', 'pandas']
PresetBackends       = Literal['json', 'yaml', 'pickle']
ContentPath          = str
Props                = Dict[str, Any]
PropsConstraints     = Dict[str, Any]
PropsList            = List[Props]


class BackendContract(ABC):
    def read(self, fs : AbstractFileSystem, path : str, constraints : Dict[str, Any]) -> PropsList:
        raise NotImplementedError("Abstract")

    def write(self, fs : AbstractFileSystem, path : str, props_list : PropsList) -> None:
        raise NotImplementedError("Abstract")


class FrontendContract(Generic[TContent, TContents], ABC):
    def read_content(self, props_list : PropsList) -> TContent:
        raise NotImplementedError("Abstract")

    def read_contents(self, props_list : PropsList) -> TContents:
        raise NotImplementedError("Abstract")

    def write_content(self, content : TContent) -> PropsList:
        raise NotImplementedError("Abstract")

    def write_contents(self, contents : TContents) -> PropsList:
        raise NotImplementedError("Abstract")


class FilocContract(Generic[TContent, TContents], ABC):
    def lock(self):
        raise NotImplementedError('Abstract')

    def lock_info(self):
        raise NotImplementedError('Abstract')

    def lock_force_release(self):
        raise NotImplementedError('Abstract')

    def invalidate_cache(self, constraints : Optional[PropsConstraints] = None, **constraints_kwargs : Props):
        raise NotImplementedError('Abstract')

    def read_content(self, constraints : Optional[PropsConstraints] = None, **constraints_kwargs : PropsConstraints) -> TContent:
        raise NotImplementedError('Abstract')

    def read_contents(self, constraints : Optional[PropsConstraints] = None, **constraints_kwargs : PropsConstraints) -> TContents:
        raise NotImplementedError('Abstract')

    def read_props_list(self, constraints : Optional[PropsConstraints] = None, **constraints_kwargs : PropsConstraints) -> PropsList:
        raise NotImplementedError('Abstract')

    def write_content(self, content : TContent, dry_run=False):
        raise NotImplementedError('Abstract')

    def write_contents(self, contents : TContents, dry_run=False):
        raise NotImplementedError('Abstract')

    def write_props_list(self, props_list : PropsList, dry_run=False):
        raise NotImplementedError('Abstract')



import socket
import threading
from filoc.core import LockException
import json
from logging import exception
import os
import shutil
import tempfile
from threading import Thread
import time
import unittest
from pathlib import Path

from filoc import filoc, filoc_json, FilocIO


# noinspection DuplicatedCode
class TestFilocConcurrency(unittest.TestCase):
    def setUp(self):
        self.maxDiff = None
        self.test_dir = tempfile.mkdtemp().replace('\\', '/')
        self.loc = filoc(self.test_dir + r'/id={id:d}/myfile.json', writable=True)

    def tearDown(self):
        shutil.rmtree(self.test_dir)

    def test_lock_block(self):
        with self.loc.lock():
            self.loc.write_content({'id' : 1, 'val' : 1})
            res = self.loc.read_content(id=1)
            print(res)

    def test_lock_info(self):
        host   = socket.gethostname()
        pid    = os.getpid()
        thread = threading.get_ident()

        self.assertIsNone(self.loc.lock_info())

        with self.loc.lock():            
            lock_info = self.loc.lock_info()
            
            self.assertEqual(lock_info['host'], host)
            self.assertEqual(lock_info['pid'], pid)
            self.assertEqual(lock_info['thread'], thread)
            
            self.loc.write_content({'id' : 1, 'val' : 1})
            res = self.loc.read_content(id=1)
            print(res)

        self.assertIsNone(self.loc.lock_info())

    def test_lock_force(self):
        self.loc.lock().__enter__()            
        self.loc.lock_force_release()
        self.assertIsNone(self.loc.lock_info())

    def test_lock_force(self):
        with self.loc.lock():
            self.loc.lock_force_release()
            self.assertIsNone(self.loc.lock_info())

    def test_lock_reenter(self):
        with self.loc.lock():
            with self.loc.lock():
                pass

    def test_lock_enter_from_other_thread(self):

        state = 0

        def wait_state_and_increment(expected_state):
            nonlocal state
            while state != expected_state:
                time.sleep(0.1)
            state += 1

        def async_lock():
            with self.loc.lock():
                wait_state_and_increment(1)
                wait_state_and_increment(4)
                    
        thread = Thread(target=async_lock)

        ### BEGIN ASYNC PLAY        
        wait_state_and_increment(0)  # state 0 --> 1
        thread.start()               # state 1 --> 2 (then lock is set)
        wait_state_and_increment(2)  # state 2 --> 3

        try:
            with self.loc.lock():
                self.fail("this line should never be called")
        except LockException:
            print("lock worked")
        finally:
            wait_state_and_increment(3)  # state 3 --> 4 (trigger lock release)
            wait_state_and_increment(5)  # state 5 --> 6 (lock released)


if __name__ == '__main__':
    unittest.main()



import os
from typing import Dict, Any

import yaml
from fsspec import AbstractFileSystem

from filoc.contract import PropsList, BackendContract
from filoc.utils import filter_and_coerce_loaded_file_content, coerce_file_content_to_write


class YamlBackend(BackendContract):
    def __init__(self, is_singleton) -> None:
        super().__init__()
        self.is_singleton = is_singleton

    def read(self, fs: AbstractFileSystem, path: str, constraints: Dict[str, Any]) -> PropsList:
        with fs.open(path) as f:
            return filter_and_coerce_loaded_file_content(path, yaml.load(f), constraints, self.is_singleton)

    def write(self, fs: AbstractFileSystem, path: str, props_list: PropsList) -> None:
        fs.makedirs(os.path.dirname(path), exist_ok=True)
        with fs.open(path, 'w') as f:
            return yaml.dump(coerce_file_content_to_write(path, props_list, self.is_singleton), f, indent=2)



import logging

from pandas import DataFrame, Series

from .contract import PropsList, TContents, TContent, FrontendContract

log = logging.getLogger('filoc')


class PandasFrontend(FrontendContract):
    def read_content(self, props_list: PropsList) -> TContent:
        return Series(props_list[0], dtype=object)  # dtype object in order to preserve int type, if Series contains only int and float values (elsewhere float wins)

    def read_contents(self, props_list: PropsList) -> TContents:
        return DataFrame(props_list)

    def write_content(self, content: TContent) -> PropsList:
        if isinstance(content, dict):
            return [content]
        elif isinstance(content, Series):
            return [content.to_dict()]
        else:
            raise ValueError(f'Expected dict or Series, got {type(content).__name__}') 

    def write_contents(self, contents: TContents) -> PropsList:
        if isinstance(contents, list):
            return contents
        elif isinstance(contents, DataFrame):
            return contents.to_dict(orient='records')
        else:
            raise ValueError(f'Expected list or DataFrame, got {type(contents).__name__}') 
